
--[[
GreenLight, a simple mod by Carys of Terenas.
Version: 0.1
Documentation: See ReadMe.txt
--]]
GREENLIGHT_MESSAGE_PREFIX = "GLITE"
GREENLIGHT_ID = "GREENLIGHT"
GREENLIGHT_VERSION = "1.0.0b"
GREENLIGHT_LOG = false
GREENLIGHT_RED = "red"
GREENLIGHT_YELLOW = "yellow"
GREENLIGHT_GREEN = "green"
GREENLIGHT_UNKNOWN = "unknown"
GREENLIGHT_STARTUP_TIME = 0
GREENLIGHT_REASON = {
   ["MANUAL"] = "MANUAL",
   ["DEAD"] = "DEAD",
   ["AFK"] = "AFK",
   ["MANA"] = "MANA",
   ["HEALTH"] = "HEALTH",
   ["EATING"] = "EATING",
   ["DRINKING"] = "DRINKING",
   ["UNSET"] = "UNSET",
   ["GHOST"] = "GHOST"
}

GREENLIGHT_REASON_PRIORITIES = {
   [GREENLIGHT_REASON.UNSET] = 2,
   [GREENLIGHT_REASON.MANA] = 3,
   [GREENLIGHT_REASON.HEALTH] = 4,
   [GREENLIGHT_REASON.DRINKING] = 5,
   [GREENLIGHT_REASON.EATING] = 6,
   [GREENLIGHT_REASON.AFK] = 7,
   [GREENLIGHT_REASON.DEAD] = 8,
   [GREENLIGHT_REASON.GHOST] = 9,
   [GREENLIGHT_REASON.MANUAL] = 10
}

GREENLIGHT_MY_STATE = {
   [GREENLIGHT_REASON.MANUAL] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.DEAD] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.GHOST] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.AFK] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.MANA] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.HEALTH] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.DRINKING] = GREENLIGHT_GREEN,
   [GREENLIGHT_REASON.EATING] = GREENLIGHT_GREEN
}

GREENLIGHT_MY_LAST_BROADCAST_STATE = GREENLIGHT_GREEN
GREENLIGHT_LAST_OVERALL_STATE = GREENLIGHT_GREEN
GREENLIGHT_STATE_TABLE = {}
GL_REPORT_NAME_INDEX = 1
GL_REPORT_STATUS_INDEX = 2
GL_REPORT_REASON_INDEX = 3
GL_REPORT_VERSION_INDEX = 4

GL_REPORT_SORT_BY_NAME=1
GL_REPORT_SORT_BY_STATUS=2
GL_REPORT_SORT_BY_REASON=3

GREENLIGHT_CONFIG = {
   ["DEFAULT_CHANNEL"] = "GLPARTY",
   ["CHANNEL"] = "GLPARTY",
   ["AUTO_USE_CTRA_CHANNEL"] = true,
   ["SORT"] = GL_REPORT_SORT_BY_STATUS,
   ["AUTO_SET_MANA"] = true,
   ["MANA_YELLOW"] = 70,
   ["MANA_RED"] = 30,
   ["AUTO_SET_HEALTH"] = true,
   ["HEALTH_YELLOW"] = 70,
   ["HEALTH_RED"] = 30
}


function GL_Print( msg ) 
   local chatframe
   chatframe = DEFAULT_CHAT_FRAME
   chatframe:AddMessage( msg );
end

function GL_Log( msg ) 
   if GREENLIGHT_LOG then
      -- output the message to console
      GL_Print( "GreenLight: Log: " .. msg );
   end
end

function GL_OnLoad()
   this:RegisterEvent("ADDON_LOADED")
   this:RegisterEvent("CHAT_MSG_CHANNEL")
   this:RegisterEvent("VARIABLES_LOADED")
   this:RegisterEvent("PARTY_MEMBERS_CHANGED")
   this:RegisterEvent("RAID_ROSTER_UPDATE")
   -- events which may affect player state
   this:RegisterEvent("UNIT_HEALTH")
   this:RegisterEvent("UNIT_MANA")
   this:RegisterEvent("UNIT_RAGE")
   this:RegisterEvent("UNIT_ENERGY")
   this:RegisterEvent("PLAYER_FLAGS_CHANGED")
   this:RegisterEvent("PLAYER_DEAD")
   this:RegisterEvent("PLAYER_UNGHOST")
   this:RegisterEvent("PLAYER_ALIVE")
   this:RegisterEvent("CHAT_MSG_SYSTEM")
   this:RegisterEvent("PLAYER_AURAS_CHANGED")
   -- slash commands
   SLASH_GREENLIGHT1 = "/greenlight"
   SLASH_GREENLIGHT2 = "/gl"
   GREENLIGHT_STARTUP_TIME = GetTime()
   SlashCmdList["GREENLIGHT"] = GL_ConsoleCommand
end


GREENLIGHT_CHANNELNUM = nil
GREENLIGHT_UPDATE_RETRY = nil
GREENLIGHT_ROSTER_CHANGED_BROADCAST_TIMER = nil
function GL_OnUpdate()
   -- a few gyrations to send a message 2 sec or so after the roster changes
   if GREENLIGHT_ROSTER_CHANGED_BROADCAST_TIMER then
      if GetTime() > GREENLIGHT_ROSTER_CHANGED_BROADCAST_TIMER then
	 GL_Log("Timer expired, broadcasting state")
	 GL_BroadcastMyState(true)
	 GREENLIGHT_ROSTER_CHANGED_BROADCAST_TIMER = nil
      end
   end

   -- this is a bunch of gyrations so that we set the channel
   -- AFTER all the other channels are loaded
   -- got this idea from GlobalComm, but I don't like latching into OnUpdate
   -- so I'll look for other ideas


   if GREENLIGHT_CHANNELNUM then return end
   local curTime = GetTime()

   if not GREENLIGHT_UPDATE_RETRY then
      GREENLIGHT_UPDATE_RETRY = curTime + 1
   end

   if GREENLIGHT_UPDATE_RETRY > curTime then
      return
   end

   --okay, we're past our update--let's see if channels are joined!
   --if there is no channel 1, don't try to connect (unless we timed out)
   local _,name = GetChannelName(1)
   if (not name) and (curTime - GREENLIGHT_STARTUP_TIME < 300) then return end
   -- okay, so we've either had the channels load or we've timed out
   GL_SetChannel({GREENLIGHT_CONFIG.CHANNEL})
   GREENLIGHT_CHANNELNUM = GetChannelName(GREENLIGHT_CONFIG.CHANNEL)
   GL_Log("GREENLIGHT_CHANNELNUM == "..GREENLIGHT_CHANNELNUM)
   if (GREENLIGHT_CHANNELNUM == 0) then
      GREENLIGHT_UPDATE_RETRY = GetTime() + 10
      GREENLIGHT_CHANNELNUM = nil
   end
end


function GL_OnRosterUpdate()
--   if (GL_IsInRaid() and GREENLIGHT_AUTO_USE_CTRA_CHANNEL and CT_RA_Channel) then
--      GREENLIGHT_CONFIG.CHANNEL = CT_RA_Channel
--   else
--      GREENLIGHT_CONFIG.CHANNEL = GREENLIGHT_CONFIG.DEFAULT_CHANNEL
--   end
   GL_RefreshStateTable()
   --GL_BroadcastMyState( true )
   GREENLIGHT_ROSTER_CHANGED_BROADCAST_TIMER = GetTime() + 2
end

GREENLIGHT_EVENT_HANDLERS = {
   ["ADDON_LOADED"] = function() GL_Log("Addon Loaded") end,
   ["VARIABLES_LOADED"] = function() 
			     GL_Log("Variables Loaded");
			     GL_RefreshStateTable()
			     GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_GREEN)
			  end,
   ["CHAT_MSG_CHANNEL"] = function() 
			     GL_Log("Received message: " .. arg2 .. ":" .. arg1); 
			     GL_ProcessChatMessage(arg2, arg9, arg1)
			  end,
   ["RAID_ROSTER_UPDATE"] = function() 
			       GL_Log("Event: " .. event); 
			       GL_OnRosterUpdate() 
			    end,
   ["PARTY_MEMBERS_CHANGED"] = function() 
			       GL_Log("Event: " .. event); 
			       GL_OnRosterUpdate() 
			    end,
   ["PLAYER_DEAD"] = function()
			GL_SetMyState(GREENLIGHT_REASON.DEAD, GREENLIGHT_RED)
		     end,
   ["PLAYER_ALIVE"] = function()
			 -- really check for death in this case, since PLAYER_ALIVE also fires when you
			 -- feign death, release and become a ghost, etc
			 if UnitIsDead("player") then
			    GL_SetMyState(GREENLIGHT_REASON.DEAD, GREENLIGHT_RED)
			 elseif UnitIsGhost("player") then
			    GL_SetMyState(GREENLIGHT_REASON.GHOST, GREENLIGHT_RED)
			 else
			    GL_SetMyState(GREENLIGHT_REASON.DEAD, GREENLIGHT_GREEN)
			    GL_SetMyState(GREENLIGHT_REASON.GHOST, GREENLIGHT_GREEN)
			 end
		      end,
   ["PLAYER_UNGHOST"] = function()
			   -- really check for death in this case, since PLAYER_ALIVE also fires when you
			   -- feign death, release and become a ghost, etc
			   if UnitIsDead("player") then
			      GL_SetMyState(GREENLIGHT_REASON.DEAD, GREENLIGHT_RED)
			   elseif UnitIsGhost("player") then
			      GL_SetMyState(GREENLIGHT_REASON.GHOST, GREENLIGHT_RED)
			   else
			      GL_SetMyState(GREENLIGHT_REASON.DEAD, GREENLIGHT_GREEN)
			      GL_SetMyState(GREENLIGHT_REASON.GHOST, GREENLIGHT_GREEN)
			   end
			end,
   ["PLAYER_FLAGS_CHANGED"] = function()
				 GL_Log("Flags Changed")
			      end,
   ["UNIT_HEALTH"] = function()
			if arg1 == "player" then
			   GL_OnHealthChanged()
			end
		     end,
   ["UNIT_MANA"] = function()
		      if arg1 == "player" then
			 GL_OnManaChanged()
		      end
		   end,
   ["UNIT_RAGE"] = function()
		      if arg1 == "player" then
			 GL_OnManaChanged()
		      end
		   end,
   ["UNIT_ENERGY"] = function()
		      if arg1 == "player" then
			 GL_OnManaChanged()
		      end
		   end,
   ["CHAT_MSG_SYSTEM"] = function()
			    local afkStr = string.gsub(MARKED_AFK_MESSAGE, "%%s", "(.-)")
			    if string.find(arg1, afkStr) then
			       GL_SetMyState(GREENLIGHT_REASON.AFK, GREENLIGHT_RED)
			    elseif arg1 == CLEARED_AFK then
			       GL_SetMyState(GREENLIGHT_REASON.AFK, GREENLIGHT_GREEN)
--			    elseif string.find(arg1, string.sub(MARKED_DND,0,-5)) then
--			       GL_Print("GL: DND w/message")
--			    elseif arg1 == CLEARED_DND then
--			       GL_Print("GL: Not DND")
			    end
			 end,
   ["PLAYER_AURAS_CHANGED"] = function()
				 if GL_IsEating() then
				    GL_SetMyState(GREENLIGHT_REASON.EATING, GREENLIGHT_RED)
				 else
				    GL_SetMyState(GREENLIGHT_REASON.EATING, GREENLIGHT_GREEN)
				 end
				 if GL_IsDrinking() then
				    GL_SetMyState(GREENLIGHT_REASON.DRINKING, GREENLIGHT_RED)
				 else
				    GL_SetMyState(GREENLIGHT_REASON.DRINKING, GREENLIGHT_GREEN)
				 end
			      end
}

function GL_BuffsMatchPattern(p)
   --thanks for hint from Gello of Underhill for this idea
   local f,b 
   for i=1,24 do 
      b=UnitBuff("player",i) or "" 
      f=f or strfind(b,p)
   end 
   if f then
      return true
   end
end

function GL_IsEating()
   return GL_BuffsMatchPattern("&Knife$")
end

function GL_IsDrinking()
   return GL_BuffsMatchPattern("\\INV_Drink")
end

function GL_WorstStatus(lhs,rhs)
   --returns whichever status is "worst" (unknown < red < yellow < green)
   if lhs == GREENLIGHT_RED or rhs == GREENLIGHT_RED then
      return GREENLIGHT_RED
   elseif lhs == GREENLIGHT_YELLOW or rhs == GREENLIGHT_YELLOW then
      return GREENLIGHT_YELLOW
   elseif lhs == GREENLIGHT_GREEN or rhs == GREENLIGHT_GREEN then
      return GREENLIGHT_GREEN
   else
      return GREENLIGHT_UNKNOWN
   end
end

function GL_IsWorseThan(lhs,rhs)
   -- returns true if the LHS argument is actually worse than the rhs argument,
   -- false if they are equal or LHS is better than RHS
   -- strangely, "unknown" is the "best" status--because we prefer having ANY status to that
   if lhs == GREENLIGHT_RED then
      if rhs == GREENLIGHT_RED then
	 return false
      else
	 return true
      end
   elseif lhs == GREENLIGHT_YELLOW then
      if rhs == GREENLIGHT_GREEN or rhs == GREENLIGHT_UNKNOWN then
	 return true
      else
	 return false
      end
   elseif lhs == GREENLIGHT_GREEN then
      if rhs == GREENLIGHT_UNKNOWN then
	 return true
      else
	 return false
      end
   end
end


function GL_OnHealthChanged()
   GL_Log("Health Changed")
   
   if GREENLIGHT_CONFIG.AUTO_SET_HEALTH then
      local health = UnitHealth("player")
      local healthmax = UnitHealthMax("player")
      local pct = (health / healthmax) * 100
      GL_Log("Player now at "..pct)
      if pct > GREENLIGHT_CONFIG.HEALTH_YELLOW then
	 GL_Log("Health Green")
	 GL_SetMyState(GREENLIGHT_REASON.HEALTH, GREENLIGHT_GREEN)
	 GREENLIGHT_MY_STATE.HEALTH = GREENLIGHT_GREEN
      elseif pct > GREENLIGHT_CONFIG.HEALTH_RED then
	 GL_Log("Health Yellow")
	 GL_SetMyState(GREENLIGHT_REASON.HEALTH, GREENLIGHT_YELLOW)
	 GREENLIGHT_MY_STATE.HEALTH = GREENLIGHT_YELLOW
      else
	 GL_Log("Health Red")
	 GL_SetMyState(GREENLIGHT_REASON.HEALTH, GREENLIGHT_RED)
	 GREENLIGHT_MY_STATE.HEALTH = GREENLIGHT_RED
      end
   else
      GREENLIGHT_MY_STATE.HEALTH = GREENLIGHT_GREEN
   end
end

function GL_OnManaChanged()
   GL_Log("Mana/Rage/Energy changed")
   -- warriors just don't watch mana; it's reverse, y'see
   local playerclass, englishclass = UnitClass("player")
   if englishclass == "WARRIOR" then
      if GREENLIGHT_MY_STATE.MANA ~= GREENLIGHT_GREEN then
	 GL_SetMyState(GREENLIGHT_REASON.MANA, GREENLIGHT_GREEN)
      end
   elseif GREENLIGHT_CONFIG.AUTO_SET_MANA then
      local mana = UnitMana("player")
      local manamax = UnitManaMax("player")
      local pct = (mana / manamax) * 100
      GL_Log("Player now at "..pct)
      if pct > GREENLIGHT_CONFIG.MANA_YELLOW then
	 GL_Log("Mana Green")
	 GL_SetMyState(GREENLIGHT_REASON.MANA, GREENLIGHT_GREEN)
	 GREENLIGHT_MY_STATE.MANA = GREENLIGHT_GREEN
      elseif pct > GREENLIGHT_CONFIG.MANA_RED then
	 GL_Log("Mana Yellow")
	 GL_SetMyState(GREENLIGHT_REASON.MANA, GREENLIGHT_YELLOW)
	 GREENLIGHT_MY_STATE.MANA = GREENLIGHT_YELLOW
      else
	 GL_Log("Mana Red")
	 GL_SetMyState(GREENLIGHT_REASON.MANA, GREENLIGHT_RED)
	 GREENLIGHT_MY_STATE.MANA = GREENLIGHT_RED
      end
   else
      GREENLIGHT_MY_STATE.MANA = GREENLIGHT_GREEN
   end
end

function GL_OnEvent(event)
   -- process our incoming events
   func = GREENLIGHT_EVENT_HANDLERS[event]
   if func then
      func()
   end
end

function GL_IsValidName( name )
   if name == UnitName("player") then
      return true
   end
   if GL_IsInGroup() then
      for k,unitName in yak_eachPartyName() do
	 if name == unitName then
	    return true
	 end
      end
      return false
   elseif GL_IsInRaid() then
      for k, unitId in yak_eachRaidID() do
	 if UnitName(unitId) == name then
	    return true
	 end
      end
      return false
   end
   return false
end
   
function GL_ChannelToUse()
   local channel = GREENLIGHT_CONFIG.CHANNEL
   if GL_IsInRaid() and GREENLIGHT_CONFIG.AUTO_USE_CTRA_CHANNEL then
      channel = CT_RA_Channel
   end
   return channel
end

function GL_ProcessChatMessage(name, channel, fullmsg)
   if channel == GL_ChannelToUse() then
      bits = GL_string_split(fullmsg)
      if bits[1] ~= GREENLIGHT_MESSAGE_PREFIX then
	 GL_Log("Not GreenLight message; discarding")
      elseif (GL_IsValidName( name )) and name ~= UnitName("player") then
	 local msg = bits[2]
	 local reason = bits[3]
	 local version = bits[4]
	 if ((msg=="red") or (msg=="yellow") or (msg=="green")) then
	    GL_Log("Changing " .. name .. " to " .. msg .. "because: "..reason)
	    GREENLIGHT_STATE_TABLE[name]={msg, reason, version}
	    GL_RecalcOverallState()
	 elseif msg == "request" then
	    GL_Log("Request received; broadcasting my state")
	    GL_BroadcastMyState( true )
	 end
      end
   end
end

function GL_ToggleLog()
   GREENLIGHT_LOG = (GREENLIGHT_LOG == false)
end

function GL_string_split(s)
   local result={}
   for match in string.gfind(s, "[^%s]+" ) do
      table.insert(result,match)
   end
   return result
end

function GL_SubCommands(s)
   local result = GL_string_split(s)
   result_command = result[1]
   table.remove(result,1)
   return result_command, result
end

function GL_HasJoinedChannel( name ) 
   for i=1, 10 do
      local channelNumber, channelName = GetChannelName(i);
      if channelName then
	 if channelName == name then
	    return true
	 end
      end
   end
end

function GL_SetChannel( args )
   if getn(args) < 1 then
      GL_Print("GreenLight currently reports on channel "..GREENLIGHT_CONFIG.CHANNEL)
   else
      local channel = args[1]
      GL_Print("GreenLight: Setting channel to "..channel)
      if GL_HasJoinedChannel( GREENLIGHT_CONFIG.CHANNEL ) and (GREENLIGHT_CONFIG.CHANNEL ~= channel) then
	 GL_Log("GreenLight: Leaving channel "..GREENLIGHT_CONFIG.CHANNEL)
	 LeaveChannelByName(GREENLIGHT_CONFIG.CHANNEL);
      end

      GREENLIGHT_CONFIG.CHANNEL = channel
      GREENLIGHT_CONFIG.DEFAULT_CHANNEL = channel
      GL_Log("Joining channel "..GREENLIGHT_CONFIG.CHANNEL)
      JoinChannelByName(GREENLIGHT_CONFIG.CHANNEL, nil, nil )
   end
end


GL_SUBCOMMANDS = {
   ["statetable"] = function () GL_PrintStateTable(GREENLIGHT_STATE_TABLE) end,
   ["status"] = function() GL_PrintStatus() end,
   ["report"] = function() GL_Report() end,
   ["log"] =  function() GL_ToggleLog() end,
   ["refresh"] = function() GL_RefreshStateTable() end,
   ["red"] = function() GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_RED) end,
   ["yellow"] = function() GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_YELLOW) end,
   ["green"] = function() GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_GREEN) end,
   ["show"] = function() GreenLightFrame:Show() end,
   ["hide"] = function() GreenLightFrame:Hide() end,
   ["config"] = function() GreenLightConfigFrame:Show() end,
   ["channel"] = function(args) GL_SetChannel(args) end, 
   ["sync"] = function() GL_BroadcastSyncRequest() end
}

function GL_BroadcastSyncRequest()
   SendChatMessage(GREENLIGHT_MESSAGE_PREFIX .. " request", "CHANNEL", "Common", GetChannelName(GL_ChannelToUse()) )
   GL_BroadcastMyState(true)
end

function GL_ConsoleCommand( message )
   -- parse and act on our console commands
   local commandlist = { }
   local command
   GL_Log("Command: " .. message)
   command, args = GL_SubCommands(message)
   if command then
      GL_Log("SubCommand: " .. command)
   end
   comfunc = GL_SUBCOMMANDS[command]
   if comfunc ~= nil then
      comfunc(args)
   else
      GL_Print("/greenlight, /gl   -- gives this usage statement")
      GL_Print("/gl red|green|blue -- sets your state")
      GL_Print("/gl refresh        -- refreshes your state table")
      GL_Print("/gl log            -- toggle display of debug messages")
      GL_Print("/gl report         -- shows current status report")
      GL_Print("/gl show|hide      -- shows/hides light frame")
      GL_Print("/gl config         -- configures GreenLight")
      GL_Print("/gl channel (name) -- sets and joins GreenLight channel")
      GL_Print("/gl sync           -- requests a sync with other clients in channel")
   end
end

function GL_SetMyState( type, color, reason )
   GREENLIGHT_MY_STATE[type] = color
   -- now calculate my new state
   local mystate
   local myreason
   mystate, myreason = GL_MyStateAndReason()
   GL_BroadcastMyState()
   GREENLIGHT_STATE_TABLE[UnitName("player")] = {mystate, myreason, GREENLIGHT_VERSION}
   -- set the button in the UI frame -- disable all first, then set ours
   GreenLightFrameButton_red:SetButtonState("NORMAL")
   GreenLightFrameButton_yellow:SetButtonState("NORMAL")
   GreenLightFrameButton_green:SetButtonState("NORMAL")
   getglobal("GreenLightFrameButton_"..GREENLIGHT_MY_STATE.MANUAL):SetButtonState("PUSHED", true)
   GL_RecalcOverallState()
end

function GL_BroadcastMyState( force )
   -- also broadcast this on the channel
   local color, reason = GL_MyStateAndReason()
   -- if not GL_HasJoinedChannel( GL_ChannelToUse() ) then
      --GL_SetChannel( {GL_ChannelToUse()} )
   -- end
   if color ~= GREENLIGHT_MY_LAST_BROADCAST_STATE or force then
      GL_Log("Broadcasting my state as " .. color .. " over channel " .. GetChannelName(GL_ChannelToUse()).. " (" .. GL_ChannelToUse() .. ")")
      SendChatMessage(GREENLIGHT_MESSAGE_PREFIX .. " " .. color.. " " .. reason .. " " .. GREENLIGHT_VERSION, "CHANNEL", "Common", GetChannelName(GL_ChannelToUse()) )
      GREENLIGHT_MY_LAST_BROADCAST_STATE = color
   end
end

function GL_IsSolo()
   return (GetNumPartyMembers() == 0) and not GL_IsInRaid()
end

function GL_IsInGroup()
   return ((not GL_IsInRaid()) and (GetNumPartyMembers() > 0))
end

function GL_IsInRaid()
   return (GetNumRaidMembers() > 0)
end

function GL_GroupSize()
   -- Gets the size of the operating group the player is
   -- a part of.  Includes the player, so if 
   -- GL_GroupSize() returns 1, the player is solo
   if GL_IsInRaid() then
      return GetNumRaidMembers()
   elseif GL_IsInParty() then
      return GetNumPartyMembers() + 1
   else
      return 1
   end
end

function GL_BlankEntriesFromSolo()
   local result = {}
   result[UnitName("player")] = {GREENLIGHT_UNKNOWN,GREENLIGHT_REASON.UNSET,GREENLIGHT_VERSION_UNSET}
   return result
end

function GL_BlankEntriesFromGroup()
   local result = {}
   local unitId
   local index
   for k, unitName in yak_eachPartyName() do
      GL_Log("Adding ".. unitName)
      result[unitName] = {GREENLIGHT_UNKNOWN,GREENLIGHT_REASON.UNSET}
   end
   result[UnitName("player")] = {GREENLIGHT_UNKNOWN,GREENLIGHT_REASON.UNSET,GREENLIGHT_VERSION_UNSET}
   --GL_PrintStateTable(result)
   return result
end

function GL_BlankEntriesFromRaid()
   local result = {}
   local name
   local index
   for k, unitId in yak_eachRaidID() do
      GL_Log("Adding "..UnitName(unitId))
      result[UnitName(unitId)] = {GREENLIGHT_UNKNOWN,GREENLIGHT_REASON.UNSET, GREENLIGHT_VERSION_UNSET}
   end
   result[UnitName("player")] = {GREENLIGHT_UNKNOWN, GREENLIGHT_REASON.UNSET, GREENLIGHT_VERSION_UNSET}
   return result
end
   
function GL_NewStateTable( oldTable )
   -- create the new table using the 
   -- names of the characters in the current group
   local result
   if GL_IsSolo() then
      result = GL_BlankEntriesFromSolo()
   elseif GL_IsInGroup() then
      result = GL_BlankEntriesFromGroup()
   elseif GL_IsInRaid() then
      result = GL_BlankEntriesFromRaid()
   end
   for key,item in GREENLIGHT_STATE_TABLE do
      if result[key] then
	 result[key] = item
      end
   end
   local mystate, myreason = GL_MyStateAndReason()
   result[UnitName("player")] = {mystate, myreason, GREENLIGHT_VERSION}
   return result
end
   
function GL_MyStateAndReason()
   -- calculates state and reason
   -- if our manual state is worse than the auto state, we use manual
   -- otherwise (ie manual == auto or auto is worse) we use auto
   local result = GREENLIGHT_GREEN
   local worstreason = GREENLIGHT_REASON.UNSET
   --first pass -- calculate worst state
   for k,v in GREENLIGHT_MY_STATE do
      if GL_IsWorseThan(v,result) then
	 result = v
      end
   end
   -- now a pass to figure out reason
   for k,v in GREENLIGHT_MY_STATE do
      if (v==result) then
	 if GREENLIGHT_REASON_PRIORITIES[k] > GREENLIGHT_REASON_PRIORITIES[worstreason] then
	    GL_Log("Replacing reason " .. worstreason .. " with " .. k)
	    worstreason = k
	 end
      end
   end
   -- if we're green, reverse that priority list!
   return result, GREENLIGHT_REASON[worstreason]
end

function GL_RefreshStateTable()
   GREENLIGHT_STATE_TABLE = GL_NewStateTable( GREENLIGHT_STATE_TABLE )
   GL_RecalcOverallState()
end


function GL_PrintMyStateTable()
   for k,v in GREENLIGHT_MY_STATE do
      GL_Print(k..":"..v)
   end
end

function GL_PrintStateTable(t)
   GL_Log("Printing state table")
   for key,item in t do
      GL_Print("Name: " .. key .. "; State: " .. item[1] .. "Reason: " .. item[2])
      if item[3] then GL_Print( "Version: " .. item[3] ) end
   end
end

function GL_Status()
   -- returns a table of the current characters' status
   -- keyed by color, with each item being the number of 
   -- players in that status
   result = {}
   result[GREENLIGHT_GREEN] = 0
   result[GREENLIGHT_RED] = 0
   result[GREENLIGHT_YELLOW] = 0
   result[GREENLIGHT_UNKNOWN] = 0
   for key,item in GREENLIGHT_STATE_TABLE do
--      GL_Print(key .. ":" ..item[1] .. ":" .. item[2])
      result[item[1]] = result[item[1]] + 1
   end
   return result
end

function GL_BoolString(f)
   if (f) then
      return "True"
   else
      return "False"
   end
end

function GL_GroupWorstStatus()
   local s = GL_Status()
   local result
   if s[GREENLIGHT_RED] > 0 then
      result = GREENLIGHT_RED;
   elseif s[GREENLIGHT_YELLOW] > 0 then
      result = GREENLIGHT_YELLOW
   else
      result = GREENLIGHT_GREEN
   end
   GL_Log( result )
   return result
end

GL_WORD_TABLE = { 
   [GREENLIGHT_RED]="Red",
   [GREENLIGHT_GREEN] = "Green",
   [GREENLIGHT_YELLOW] = "Yellow",
   [GREENLIGHT_UNKNOWN] = "Unknown"
}
   

function GL_StatusWord(arg)
   return GL_WORD_TABLE[arg]
end

function GL_PrintStatus()
   local s = GL_Status()
   GL_Print("Overall status: " .. GL_StatusWord(GL_GroupWorstStatus()) )
   GL_Print("G:"..s[GREENLIGHT_GREEN].."/Y:"..s[GREENLIGHT_YELLOW].."/R:"..s[GREENLIGHT_RED].."/?:"..s[GREENLIGHT_UNKNOWN])
end

function GL_BrightOrDark( state, color )
   if (color == state) then
      GL_Log("bright"..color)
      return "bright"
   else
      GL_Log("dark"..color)
      return "dark"
   end
end

function GL_RecalcOverallState()
   local new_state
   GL_Log("Calculating new state")
   new_state = GL_GroupWorstStatus()
   if new_state ~= GREENLIGHT_LAST_OVERALL_STATE then
      GL_Log("STATE CHANGE from " .. GREENLIGHT_LAST_OVERALL_STATE .. " to " .. new_state )
      -- set indicators
      GreenLightFrameButton_redIndicator:SetTexture("Interface\\Addons\\GreenLight\\Images\\red" .. GL_BrightOrDark(new_state, GREENLIGHT_RED))
      GreenLightFrameButton_yellowIndicator:SetTexture("Interface\\Addons\\GreenLight\\Images\\yellow" .. GL_BrightOrDark(new_state, GREENLIGHT_YELLOW))
      GreenLightFrameButton_greenIndicator:SetTexture("Interface\\Addons\\GreenLight\\Images\\green" .. GL_BrightOrDark(new_state, GREENLIGHT_GREEN))
      getglobal("GreenLightFrameButton_"..GREENLIGHT_LAST_OVERALL_STATE):SetNormalTexture("Interface\\Addons\\Greenlight\\Images\\buttonunpushed")
      getglobal("GreenLightFrameButton_"..new_state):SetNormalTexture("Interface\\Addons\\Greenlight\\Images\\buttonauto")
      -- play sound
      GREENLIGHT_LAST_OVERALL_STATE = new_state
   end
   -- update indicator strings
   local status = GL_Status()
   GreenLightFrameUnknownIndicatorString:SetText(status[GREENLIGHT_UNKNOWN])
   GreenLightFrameButton_redIndicatorString:SetText(status[GREENLIGHT_RED])
   GreenLightFrameButton_yellowIndicatorString:SetText(status[GREENLIGHT_YELLOW])
   GreenLightFrameButton_greenIndicatorString:SetText(status[GREENLIGHT_GREEN])
   
end



function GL_Frame_OnLoad()
   this:RegisterForDrag("LeftButton")
end

function GL_Frame_OnDragStart()
   GreenLightFrame:StartMoving()
end

function GL_Frame_OnDragStop()
   GreenLightFrame:StopMovingOrSizing()
end

function GL_PopulateReportTable()
   GREENLIGHT_REPORT_TABLE={}
   for k,v in GREENLIGHT_STATE_TABLE do
      GL_Log("Adding " .. k .. ":" .. v[1] .. ":" .. v[2]..":")
      table.insert(GREENLIGHT_REPORT_TABLE, {k,v[1],GREENLIGHT_REASON_STRING[v[2]],v[3]})
   end
end

function GL_RefreshReport()
   GL_PopulateReportTable()
   GL_ScrollReportUpdate()
end

function GL_Report()
   GL_PopulateReportTable()
   GL_ScrollReportUpdate()
   GreenLightReportFrame:Show()
end

GREENLIGHT_COLOR_TABLE = {
   [GREENLIGHT_RED] = {1,0,0},
   [GREENLIGHT_YELLOW] = {1,1,0},
   [GREENLIGHT_GREEN] = {0,1,0},
   [GREENLIGHT_UNKNOWN] = {1,1,1}
}

function GL_ScrollReportUpdate()
   local numEntries = getn(GREENLIGHT_REPORT_TABLE);
   FauxScrollFrame_Update(GreenLightReportFrameScrollFrame, numEntries, 19, 20);

   for i = 1, 19, 1 do
      local button = getglobal("GreenLightReportFramePlayer" .. i);
      local index = i + FauxScrollFrame_GetOffset(GreenLightReportFrameScrollFrame);
      if ( index <= numEntries ) then
	 button:Show();
	 local color = GREENLIGHT_COLOR_TABLE[GREENLIGHT_REPORT_TABLE[index][GL_REPORT_STATUS_INDEX]]
	 if ( color ) then
	    getglobal(button:GetName() .. "Status"):SetTextColor(color[1], color[2], color[3]);
	    getglobal(button:GetName() .. "Reason"):SetTextColor(color[1], color[2], color[3]);
	 end
	 getglobal(button:GetName() .. "Name"):SetText(GREENLIGHT_REPORT_TABLE[index][GL_REPORT_NAME_INDEX])
	 getglobal(button:GetName() .. "Status"):SetText(GREENLIGHT_REPORT_TABLE[index][GL_REPORT_STATUS_INDEX])
	 getglobal(button:GetName() .. "Reason"):SetText(GREENLIGHT_REPORT_TABLE[index][GL_REPORT_REASON_INDEX])
	 getglobal(button:GetName() .. "Version"):SetText(GREENLIGHT_REPORT_TABLE[index][GL_REPORT_VERSION_INDEX])
      else
	 button:Hide();
      end
   end

end

GREENLIGHT_LT = 1
GREENLIGHT_EQ = 2
GREENLIGHT_GT = 3

GREENLIGHT_STATUS_LESS_THAN_TABLE = {
   [GREENLIGHT_UNKNOWN] = { 
      [GREENLIGHT_UNKNOWN] = GREENLIGHT_EQ,
      [GREENLIGHT_RED] = GREENLIGHT_GT,
      [GREENLIGHT_YELLOW] = GREENLIGHT_GT,
      [GREENLIGHT_GREEN] = GREENLIGHT_GT
   },
   [GREENLIGHT_RED] = {
      [GREENLIGHT_UNKNOWN] = GREENLIGHT_LT,
      [GREENLIGHT_RED] = GREENLIGHT_EQ,
      [GREENLIGHT_YELLOW] = GREENLIGHT_LT,
      [GREENLIGHT_GREEN] = GREENLIGHT_LT
   },
   [GREENLIGHT_YELLOW] = {
      [GREENLIGHT_UNKNOWN] = GREENLIGHT_LT,
      [GREENLIGHT_RED] = GREENLIGHT_GT,
      [GREENLIGHT_YELLOW] = GREENLIGHT_EQ,
      [GREENLIGHT_GREEN] = GREENLIGHT_LT
   },
   [GREENLIGHT_GREEN] = {
      [GREENLIGHT_UNKNOWN] = GREENLIGHT_LT,
      [GREENLIGHT_RED] = GREENLIGHT_GT,
      [GREENLIGHT_YELLOW] = GREENLIGHT_GT,
      [GREENLIGHT_GREEN] = GREENLIGHT_EQ
   }
}
      

GREENLIGHT_SORT_FUNCTIONS = {
   function (a,b) 
      return a[GL_REPORT_NAME_INDEX] < b[GL_REPORT_NAME_INDEX] 
   end,
   function (a,b) 
      local result = GREENLIGHT_STATUS_LESS_THAN_TABLE[a[GL_REPORT_STATUS_INDEX]][b[GL_REPORT_STATUS_INDEX]]
      if result == GREENLIGHT_EQ then
	 result = a[GL_REPORT_NAME_INDEX] < b[GL_REPORT_NAME_INDEX]
      else
	 if result == GREENLIGHT_LT then
	    result = true
	 else
	    result = false
	 end
      end
      return result
   end,
--      local av
--      local bv
--      bv = b[GL_REPORT_STATUS_INDEX]
--      av  = a[GL_REPORT_STATUS_INDEX]
--      if av == GREENLIGHT_RED then
--	 if b ~= GREENLIGHT_RED then
--	    return true
--	 end
--      elseif av == GREENLIGHT_YELLOW then
--	 if bv == GREENLIGHT_RED then
--	    return false
--	 elseif bv == GREENLIGHT_GREEN then
--	    return true
--	 end
--      else -- av == green
--	 if bv ~= GREENLIGHT_GREEN  then
--	    return false
--	 end
--      end
--   end,
   function (a,b) 
      local result =  a[GL_REPORT_REASON_INDEX] < b[GL_REPORT_REASON_INDEX]
      if result == false then
	 if a[GL_REPORT_REASON_INDEX] == b[GL_REPORT_REASON_INDEX] then
	    result = a[GL_REPORT_NAME_INDEX] < b[GL_REPORT_NAME_INDEX]
	 end
      end
      return result
   end
}

function GL_SortReportByColumn(col)
   table.sort( GREENLIGHT_REPORT_TABLE, GREENLIGHT_SORT_FUNCTIONS[col] )
   GL_ScrollReportUpdate()
end

function GL_SortReportByName()
   GL_Print("Sorting by Name")
   GL_SortReportByColumn(GL_REPORT_NAME_INDEX)
end

function GL_SortReportByStatus()
   GL_Print("Sorting by Status")
   GL_SortReportByColumn(GL_REPORT_STATUS_INDEX)
end

function GL_SortReportByReason()
   GL_Print("Sorting by Reason")
   GL_SortReportByColumn(GL_REPORT_REASON_INDEX)
end

function GL_OnDragManaYellow()
   GREENLIGHT_CONFIG.MANA_YELLOW = this:GetValue();
   if (GREENLIGHT_CONFIG.MANA_YELLOW < GREENLIGHT_CONFIG.MANA_RED) then
      GreenLightConfigFrameManaRedSlider:SetValue(GREENLIGHT_CONFIG.MANA_YELLOW)
   end
   getglobal(this:GetName().."Text"):SetText(GREENLIGHT_CONFIG_MANA_YELLOW_SLIDER .. ":"..this:GetValue().."%");
end

function GL_OnDragManaRed()
   GREENLIGHT_CONFIG.MANA_RED = this:GetValue();
   if ( GREENLIGHT_CONFIG.MANA_YELLOW < GREENLIGHT_CONFIG.MANA_RED) then
      GreenLightConfigFrameManaYellowSlider:SetValue(GREENLIGHT_CONFIG.MANA_RED)
   end
   getglobal(this:GetName().."Text"):SetText(GREENLIGHT_CONFIG_MANA_RED_SLIDER .. ":"..this:GetValue().."%");
end

function GL_OnDragHealthYellow()
   GREENLIGHT_CONFIG.HEALTH_YELLOW = this:GetValue();
   if (GREENLIGHT_CONFIG.HEALTH_YELLOW < GREENLIGHT_CONFIG.HEALTH_RED) then
      GreenLightConfigFrameHealthRedSlider:SetValue(GREENLIGHT_CONFIG.HEALTH_YELLOW)
   end
   getglobal(this:GetName().."Text"):SetText(GREENLIGHT_CONFIG_HEALTH_YELLOW_SLIDER .. ":"..this:GetValue().."%");
end

function GL_OnDragHealthRed()
   GREENLIGHT_CONFIG.HEALTH_RED = this:GetValue();
   if ( GREENLIGHT_CONFIG.HEALTH_YELLOW < GREENLIGHT_CONFIG.HEALTH_RED) then
      GreenLightConfigFrameHealthYellowSlider:SetValue(GREENLIGHT_CONFIG.HEALTH_RED)
   end
   getglobal(this:GetName().."Text"):SetText(GREENLIGHT_CONFIG_HEALTH_RED_SLIDER .. ":"..this:GetValue().."%");
end

function GL_SetRed()
   GL_Log("Setting state RED")
   GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_RED)
end

function GL_SetYellow()
   GL_Log("Setting state YELLOW")
   GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_YELLOW)
end

function GL_SetGreen()
   GL_Log("Setting state GREEN")
   GL_SetMyState(GREENLIGHT_REASON.MANUAL, GREENLIGHT_GREEN)
end

function GL_ShowTooltip( color )
   GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT"); 
   GameTooltip:ClearLines()
   -- get a list of the folks who are this color
   local charlist = {}
   for k,v in GREENLIGHT_STATE_TABLE do
      if v[1] == color then
	 table.insert(charlist,{k,GREENLIGHT_REASON_STRING[v[2]]})
      end
   end
   for k,v in ipairs( charlist ) do
      local colorvals = GREENLIGHT_COLOR_TABLE[color]
      GameTooltip:AddDoubleLine(v[1], v[2], 1,1,1, colorvals[1], colorvals[2], colorvals[3])
   end
   GameTooltip:Show()
end

function GL_HideTooltip()
   GameTooltip:Hide()
end